@@ -1,4 +1,4 @@
-# Copyright 2018, 2019 Blair Bonnett
+# Copyright 2018, 2019, 2020 Blair Bonnett
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are met:
@@ -36,7 +36,7 @@ consistent-looking plots.
 
 """
 
-__version__ = "1.2.0"
+__version__ = "1.4.0"
 
 # We don't import Matplotlib here as this brings in NumPy. In turn, NumPy
 # caches a reference to the io.open() method as part of its data loading
@@ -48,7 +48,9 @@ __version__ = "1.2.0"
 # different functions.
 
 import configparser
+import importlib
 import inspect
+import io
 import os
 import os.path
 import re
@@ -334,6 +336,44 @@ class PgfutilsParser(configparser.ConfigParser):
         raise ColorError("{}.{}: could not interpret '{}' as a color.".format(section, option, value))
 
 
+    def in_tracking_dir(self, type, fn):
+        """Check if a file is in a tracking directory.
+
+        Parameters
+        ----------
+        type : {"data", "import"}
+            The type of file to check for.
+        fn : path-like
+            The filename to check.
+
+        Returns
+        -------
+        Boolean
+            True if the file is in one of the corresponding tracking
+            directories specified in the configuration.
+
+        """
+        if type == "data":
+            paths = self.get("paths", "data").strip().splitlines()
+        elif type == "import":
+            paths = self.get("paths", "pythonpath").strip().splitlines()
+            paths.extend(self.get("paths", "extra_imports").strip().splitlines())
+        else:
+            raise ValueError("Unknown tracking type {0:s}.".format(type))
+
+        # If the filename relative to one of these paths does not have to leave
+        # the directory (i.e., doesn't start with ..) then it must be within
+        # the directory.
+        fn = os.path.abspath(fn)
+        for path in paths:
+            path = os.path.abspath(path)
+            if not os.path.relpath(fn, path).startswith(".."):
+                return True
+
+        # Not in any of the directories.
+        return False
+
+
 # The current configuration.
 _config = PgfutilsParser()
 
@@ -345,48 +385,77 @@ def _config_reset():
     _config.clear()
     _config.read_dict({
         'tex': {
-           'engine': 'xelatex',
-           'text_width': '345 points',
-           'text_height': '550 points',
-           'marginpar_width': '65 points',
-           'marginpar_sep': '11 points',
-           'num_columns': '1',
-           'columnsep': '10 points',
+            'engine': 'xelatex',
+            'text_width': '345 points',
+            'text_height': '550 points',
+            'marginpar_width': '65 points',
+            'marginpar_sep': '11 points',
+            'num_columns': '1',
+            'columnsep': '10 points',
         },
 
         'pgfutils': {
-           'preamble': '',
-           'preamble_substitute': 'false',
-           'font_family': 'serif',
-           'font_name': '',
-           'font_size': '10',
-           'legend_font_size': '10',
-           'line_width': '1',
-           'axes_line_width': '0.6',
-           'legend_border_width': '0.6',
-           'legend_border_color': '(0.8, 0.8, 0.8)',
-           'legend_background': '(1, 1, 1)',
-           'legend_opacity': 0.8,
-           'figure_background': '',
-           'axes_background': 'white',
-       },
+            'preamble': '',
+            'preamble_substitute': 'false',
+            'font_family': 'serif',
+            'font_name': '',
+            'font_size': '10',
+            'legend_font_size': '10',
+            'line_width': '1',
+            'axes_line_width': '0.6',
+            'legend_border_width': '0.6',
+            'legend_border_color': '(0.8, 0.8, 0.8)',
+            'legend_background': '(1, 1, 1)',
+            'legend_opacity': 0.8,
+            'figure_background': '',
+            'axes_background': 'white',
+            'extra_tracking': '',
+        },
 
-       'rcParams': {
-       },
+        'paths': {
+            'data': '.',
+            'pythonpath': '',
+            'extra_imports': '',
+        },
 
-       'postprocessing': {
-           'fix_raster_paths': 'true',
-           'tikzpicture': 'false',
-       }
+        'rcParams': {
+        },
+
+        'postprocessing': {
+            'fix_raster_paths': 'true',
+            'tikzpicture': 'false',
+        }
     })
 
 
+def _relative_if_subdir(fn):
+    """Internal: get a relative or absolute path as appropriate.
+
+    Parameters
+    ----------
+    fn : path-like
+        A path to convert.
+
+    Returns
+    -------
+    fn : str
+        A relative path if the file is underneath the top-level project
+        directory, or an absolute path otherwise.
+
+    """
+    rel = os.path.relpath(fn)
+    if rel.startswith(".."):
+        return os.path.abspath(fn)
+    return rel
+
+
 def _file_tracker(to_wrap):
     """Internal: install an opened file tracker.
 
-    To be trackable, the given callable should return some object with `name`
-    and `mode` attributes representing the filename and mode of the opened
-    file.
+    To be trackable, the given callable should return an instance of a subclass
+    of `io.IOBase`. If its `writable` method returns True, it is assumed to be
+    an output file, otherwise if its `readable` method returns True it is
+    assumed to be an input.
 
     The docstring, name etc. of the given callable are copied to the wrapper
     function that performs the tracking.
@@ -405,14 +474,16 @@ def _file_tracker(to_wrap):
     Wrapper function.
 
     """
+    global _config
+
     import functools
     @functools.wraps(to_wrap)
     def wrapper(*args, **kwargs):
         # Defer opening to the real function.
         file = to_wrap(*args, **kwargs)
 
-        # Standard IO readers store what we need in name and mode attributes.
-        if not hasattr(file, 'name') or not hasattr(file, 'mode'):
+        # Only attempt to track files implemented as standard IO objects.
+        if not isinstance(file, io.IOBase):
             return file
 
         # Integers indicate file objects that were created from descriptors
@@ -424,21 +495,17 @@ def _file_tracker(to_wrap):
         if isinstance(file.name, int):
             return file
 
-        # Everything we are interested in is within the project tree.
-        relname = os.path.relpath(file.name)
-        if relname.startswith('.'):
-            return file
+        # If a file is writeable; this includes files opened in r+ or append
+        # modes. We assume these can't be dependencies.
+        if file.writable():
+            # Does it match the filename pattern used by the PGF backend for
+            # rasterised parts of the image being saved as PNGs?
+            if re.match(r"^.+-img\d+.png$", file.name):
+                _file_tracker.filenames.add(("w", _relative_if_subdir(file.name)))
 
-        # Assume files that are read are dependencies.
-        if 'r' in file.mode:
-            _file_tracker.filenames.add(("r", relname))
-
-        # Binary files being written could be rasterised parts of the image
-        # being saved as PNGs. Confirm this by checking the filename against
-        # the pattern that the PGF backend uses.
-        elif file.mode == 'wb':
-            if re.match(r"^.+-img\d+.png$", relname):
-                _file_tracker.filenames.add(("w", relname))
+        # Should always be readable in this case, but check anyway.
+        elif file.readable() and _config.in_tracking_dir("data", file.name):
+            _file_tracker.filenames.add(("r", _relative_if_subdir(file.name)))
 
         # Done.
         return file
@@ -450,8 +517,46 @@ def _file_tracker(to_wrap):
 _file_tracker.filenames = set()
 
 
-def _install_file_trackers():
-    """Internal: install file trackers in likely locations.
+class ImportTracker(importlib.abc.MetaPathFinder):
+    """Import finder which tracks imported files in configured paths.
+
+    """
+    def __init__(self):
+        super().__init__()
+        self._avoid_recursion = set()
+        self.imported = set()
+
+    def find_spec(self, fullname, path, target=None):
+        # According to PEP451, this is mostly intended for a reload. I can't
+        # see a way (without calling importlib._bootstrap._find_spec, which
+        # should not be imported according to the note at the top of the
+        # module) to pass this information on. Hence, lets skip tracking in
+        # this case.
+        if target is not None:  # pragma: no cover
+            return None
+
+        # We use importlib to find the actual spec, so we need to avoid
+        # recursing when this finder is called again.
+        if fullname in self._avoid_recursion:
+            return None
+
+        # Find the spec.
+        self._avoid_recursion.add(fullname)
+        spec = importlib.util._find_spec_from_path(fullname, path)
+        self._avoid_recursion.remove(fullname)
+
+        # If it has an origin in one of our tracked dirs, log it.
+        if spec is not None and spec.origin is not None:
+            global _config
+            if _config.in_tracking_dir("import", spec.origin):
+                _file_tracker.filenames.add(("r", _relative_if_subdir(spec.origin)))
+
+        # And return the result.
+        return spec
+
+
+def _install_standard_file_trackers():
+    """Internal: install standard file trackers in likely locations.
 
     This wraps the standard open() function as well as the io.open() function.
     This seems to catch all uses of the standard NumPy `load` (for both .npy
@@ -469,6 +574,73 @@ def _install_file_trackers():
     import io
     builtins.open = _file_tracker(builtins.open)
     io.open = _file_tracker(io.open)
+    sys.meta_path.insert(0, ImportTracker())
+
+
+def _install_extra_file_trackers(trackers):
+    """Internal: install requested extra file trackers.
+
+    Parameters
+    ----------
+    trackers : list
+        List of strings containing the names of the extra trackers to install.
+        Names are not case-sensitive. Currently only "netCDF4" is supported.
+
+    """
+    for tracker in trackers:
+        tracker = tracker.strip().lower()
+
+        # netCDF4 data storage.
+        if tracker == "netcdf4":
+            import netCDF4
+
+            # Wrap the Dataset class to modify its initialiser to track read
+            # files. The class is part of the compiled extension so we can't
+            # just override the initialiser.
+            class PgfutilsTrackedDataset(netCDF4.Dataset):
+                def __init__(self, filename, mode="r", **kwargs):
+                    super().__init__(filename, mode=mode, **kwargs)
+                    if mode == "r":
+                        _file_tracker.filenames.add(("r", filename))
+            netCDF4.Dataset = PgfutilsTrackedDataset
+
+            # Same deal for the MFDataset (multiple files read as one dataset).
+            class PgfutilsTrackedMFDataset(netCDF4.MFDataset):
+                def __init__(self, files, *args, **kwargs):
+                    super().__init__(files, *args, **kwargs)
+
+                    # Single string: glob pattern to expand.
+                    if isinstance(files, str):
+                        import glob
+                        files = glob.glob(files)
+
+                    # And track them all.
+                    for fn in files:
+                        _file_tracker.filenames.add(("r", fn))
+            netCDF4.MFDataset = PgfutilsTrackedMFDataset
+
+        else:
+            raise ValueError("Unknown extra tracker {}".format(tracker))
+
+
+def add_dependencies(*args):
+    """Add a file dependency for the current figure.
+
+    In most situations, the in-built file tracking should suffice. However,
+    some files can be missed (for example, if a library you are using has its
+    file opening routines in a compiled extension rather than using Python
+    functions). This function can be used to manually add these files to the
+    dependencies of the current figure.
+
+    Parameters
+    ----------
+    fn : string
+        The filename of the dependency, relative to the top-level directory of
+        the project.
+
+    """
+    for fn in args:
+        _file_tracker.filenames.add(("r", fn))
 
 
 def _list_opened_files():
@@ -536,7 +708,7 @@ def setup_figure(width=1.0, height=1.0, columns=None, margin=False,
     # Need to install our file trackers (if desired)
     # before we import Matplotlib.
     if 'PGFUTILS_TRACK_FILES' in os.environ:
-        _install_file_trackers()
+        _install_standard_file_trackers()
 
     # Reset the configuration.
     _config_reset()
@@ -549,6 +721,12 @@ def setup_figure(width=1.0, height=1.0, columns=None, margin=False,
     if kwargs:
         _config.read_kwargs(**kwargs)
 
+    # Now we can add any extra trackers specified in the config.
+    if 'PGFUTILS_TRACK_FILES' in os.environ:
+        extra = _config['pgfutils']['extra_tracking'].strip()
+        if extra:
+            _install_extra_file_trackers(extra.split(","))
+
     # Reset our interactive flag on each call.
     _interactive = False
 
@@ -564,6 +742,12 @@ def setup_figure(width=1.0, height=1.0, columns=None, margin=False,
         if ipython.active_eventloop is not None:            # pragma: no cover
             _interactive = True
 
+    # Add any desired entries to sys.path.
+    for newpath in _config['paths']['pythonpath'].splitlines():
+        newpath = newpath.strip()
+        if newpath:
+            sys.path.insert(0, newpath)
+
     # We're now ready to start configuring Matplotlib.
     import matplotlib
 
@@ -582,7 +766,7 @@ def setup_figure(width=1.0, height=1.0, columns=None, margin=False,
     # https://github.com/matplotlib/matplotlib/pull/12805
     try:
         from matplotlib.rcsetup import _validate_tex_preamble
-    except ImportError:
+    except ImportError:  # pragma: no cover
         matplotlib.rcParams.validate['pgf.preamble'] = str.splitlines
 
     # Specify which TeX engine we are using.
